#include <math.h>
#include "Graphics.h"

#define PI 3.141592654

static struct Matrix CurrentMatrix;
static struct Matrix MatrixStack[20];
static unsigned char MatrixStackPointer;
static int SizeX;
static int SizeY;

void InitGraphics( int DisplaySizeX, int DisplaySizeY )
{
  UnitMatrix( &CurrentMatrix );
  SizeX = DisplaySizeX / 2;
  SizeY = DisplaySizeY / 2;
}

void UnitMatrix( struct Matrix* pMatrix )
{
  unsigned char i, j;

  for( i = 0; i < 4; i++ )
    for( j = 0; j < 4; j++ )
      pMatrix->Werte[i][j] = 0.0;

  pMatrix->Werte[0][0] = pMatrix->Werte[1][1] = 1.0;
  pMatrix->Werte[2][2] = pMatrix->Werte[3][3] = 1.0;
}

void Apply( struct Matrix* pMatrix )
{
  unsigned char i, j, k;
  struct Matrix Help;

  for( i = 0; i < 4; i++ ) {
    for( j = 0; j < 4; j++ ) {
      Help.Werte[j][i] = 0;
      for( k = 0; k < 4; k++ ) {
        Help.Werte[j][i] += pMatrix->Werte[j][k] * CurrentMatrix.Werte[k][i];
      }
    }
  }
  
  CurrentMatrix = Help;
}

void Translate( REAL X, REAL Y, REAL Z )
{
  struct Matrix Help;

  UnitMatrix( &Help );
  Help.Werte[3][0] = X;
  Help.Werte[3][1] = Y;
  Help.Werte[3][2] = Z;

  Apply( &Help );
}

void Rotate( REAL Angle, unsigned char Axis )
{
  struct Matrix Help;
  REAL cosa, sina;

  Angle = Angle * PI / 180.0;

  cosa = (REAL)cos( Angle );
  sina = (REAL)sin( Angle );

  UnitMatrix( &Help );

  switch( Axis ) {
    case X_AXIS:
       Help.Werte[1][1] = cosa;
	     Help.Werte[2][2] = cosa;
	     Help.Werte[2][1] = sina;
	     Help.Werte[1][2] = -sina;
	     break;

    case Y_AXIS:
       Help.Werte[0][0] = cosa;
	     Help.Werte[2][2] = cosa;
	     Help.Werte[2][0] = -sina;
	     Help.Werte[0][2] = sina;
	     break;

    case Z_AXIS:
       Help.Werte[0][0] = cosa;
	     Help.Werte[1][1] = cosa;
	     Help.Werte[1][0] = sina;
	     Help.Werte[0][1] = -sina;
	     break;
  }

  Apply( &Help );
}

void Scale( REAL xs, REAL ys, REAL zs )
{
  struct Matrix Help;

  UnitMatrix( &Help );
  Help.Werte[0][0] = xs;
  Help.Werte[1][1] = ys;
  Help.Werte[2][2] = zs;
  Apply( &Help );
}

void Persp( REAL Alpha, REAL ZNear, REAL ZFar )
{
  struct Matrix Help;
  REAL cosa, sina, q;

  UnitMatrix( &Help );

  Alpha = Alpha * PI / 180.0;
  cosa = cos( Alpha / 2.0 );
  sina = sin( Alpha / 2.0 );
  q = sina / (1 - (ZNear/ZFar));

  Help.Werte[0][0] = cosa;
  Help.Werte[1][1] = cosa;
  Help.Werte[2][2] = q;
  Help.Werte[2][3] = sina;
  Help.Werte[3][2] = -q * ZNear;
  Help.Werte[3][3] = 0;

  Apply( &Help );
}

void Push()
{
  MatrixStack[MatrixStackPointer++] = CurrentMatrix;
}

void Pop()
{
  CurrentMatrix = MatrixStack[--MatrixStackPointer];
}

void Transform( struct Point* World, struct HomogenPoint* Transformed )
{
  Transformed->X =
           World->X * CurrentMatrix.Werte[0][0] +
           World->Y * CurrentMatrix.Werte[1][0] +
           World->Z * CurrentMatrix.Werte[2][0] +
           CurrentMatrix.Werte[3][0];

  Transformed->Y =
           World->X * CurrentMatrix.Werte[0][1] +
           World->Y * CurrentMatrix.Werte[1][1] +
           World->Z * CurrentMatrix.Werte[2][1] +
           CurrentMatrix.Werte[3][1];

  Transformed->Z =
           World->X * CurrentMatrix.Werte[0][2] +
           World->Y * CurrentMatrix.Werte[1][2] +
           World->Z * CurrentMatrix.Werte[2][2] +
           CurrentMatrix.Werte[3][2];

  Transformed->W =
           World->X * CurrentMatrix.Werte[0][3] +
           World->Y * CurrentMatrix.Werte[1][3] +
           World->Z * CurrentMatrix.Werte[2][3] +
           CurrentMatrix.Werte[3][3];
}

void DrawLine3d( CDC* pDC,
                 double StartX, double StartY, double StartZ,
                 double EndX, double EndY, double EndZ )
{
  struct Point Start = { StartX, StartY, StartZ };
  struct Point End = { EndX, EndY, EndZ };
  
  DrawLine( pDC, &Start, &End );
}
                 
void DrawLine( CDC* pDC, struct Point* pStart, struct Point* pEnd )
{
  struct HomogenPoint s, e;
  
  Transform( pStart, &s );
  Transform( pEnd, &e );
  
  int StartX =  ( s.X / s.W ) * SizeX + SizeX;
  int StartY = -( s.Y / s.W ) * SizeY + SizeY;
  int EndX   =  ( e.X / e.W ) * SizeX + SizeX;
  int EndY   = -( e.Y / e.W ) * SizeY + SizeY;

  lcd_line( StartX, StartY, EndX, EndY );
}


